;; 
;;  uim-keymap.el - key table for uim.el
;;
;;  Copyright (c) 2005, Konosuke Watanabe <nosuke@csc.ne.jp>
;;  All rights reserved.
;;
;;  $Id: uim-keymap.el,v 1.3 2005/07/23 13:44:59 nosuke Exp $
;;
;;  Redistribution and use in source and binary forms, with or
;;  without modification, are permitted provided that the
;;  following conditions are met:
;;
;;  1. Redistributions of source code must retain the above
;;     copyright notice, this list of conditions and the
;;     following disclaimer.
;;  2. Redistributions in binary form must reproduce the above
;;     copyright notice, this list of conditions and the
;;     following disclaimer in the documentation and/or other
;;     materials provided with the distribution.
;;  3. Neither the name of authors nor the names of its
;;     contributors may be used to endorse or promote products
;;     derived from this software without specific prior written
;;     permission.
;;
;;  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
;;  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
;;  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
;;  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;;  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
;;  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
;;  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
;;  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;;  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
;;  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
;;  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;

(defconst uim-special-keys
  '(
     backspace
     tab
     return
     escape
     prior
     next
     end
     home
     left
     up
     right
     down
     insert
     delete
     muhenkan
     henkan
     zenkaku-hankaku
     space ; for XEmacs?
     f1
     f2
     f3
     f4
     f5
     f6
     f7
     f8
     f9
     f10
     f11
     f12
     f13
     f14
     f15
     f16
     f17
     f18
     f19
     f20
     f21
     f22
     f23
     f24
     f25
     f26
     f27
     f28
     f29
     f30
     f31
     f32
     f33
     f34
     f35
     ;;; ʲϥץꥨǥåѤʤȤʤ褦Ƴ
     select
     print
     execute
     backtab
     insert
     undo
     redo
     clearline
     insertline
     deleteline
     insertchar
     deletechar
     kp-add
     kp-subtract
     kp-multiply
     kp-divide
     kp-backtab
     kp-space
     kp-tab
     kp-enter
     kp-separator
     kp-decimal
     kp-equal
     ;;;
     kp-0
     kp-1
     kp-2
     kp-3
     kp-4
     kp-5
     kp-6
     kp-7
     kp-8
     kp-9
     kp-f1
     kp-f2
     kp-f3
     kp-f4
     key-20
     scroll-lock
     ))

   

;;; metaΤ̥ޥåפ
(defconst uim-mod-keys
  '(
    ()
    (shift)
    (control)
    (shift control)
    ))



(defun uim-bind-special-keys ()
  (let ((mod-keys uim-mod-keys) 
	special-keys spkey modkey vec meta-vec)

    (while mod-keys
      ;; (wile (setq modkey (car mod-keys)) Ƚ񤯤Ȥޤ 
      ;;;   (ǽ餬nilȤߤʤ礬뤿)
      (setq modkey (car mod-keys)) 

      (setq special-keys uim-special-keys)

      (while (setq spkey (car special-keys))

	;; vec  [(shift control f1)] ߤʴ
	(setq vec (vector (append modkey (list spkey)))) 

	;; meta-vec  [(shift control meta f1)] ߤʴ
	(setq meta-vec (vector (append modkey '(meta) (list spkey))))

	;; uim-mode-mapФ(meta̵)
	;;  f1Τ褦ˡޥåפޥåפƤ륭ˤϿʤ
	;;  ñ Escape˴ؤƤꤷʤ
	(if (not (or (keymapp (key-binding vec)) 
		     (and (= (length modkey) 0)
			  (equal spkey 'escape)))) 
	    (define-key uim-mode-map  vec 'uim-process-input))

	;; uim-mode-mapФ(meta)
	;;  f1Τ褦ˡޥåפޥåפƤ륭ˤϿʤ
	;;  Meta-escapeϡXEmacsǤꤷʤ
	(if (not (or (keymapp (key-binding meta-vec))
		     (and uim-xemacs
			  (= (length modkey) 0)
			  (equal spkey 'escape))))
	    (define-key uim-mode-map meta-vec 'uim-process-input))


	;; uim-preedit-mapФ
	;;  XEmacsξ硤ñEscapeϺǸꤹΤǤǤФ
	(if (not (and uim-xemacs
		      (equal spkey 'escape)))
	    (define-key uim-preedit-map  vec 'uim-process-input))


	;; uim-preedit-mapФ(meta)
	;;  XEmacsξ硤Meta-escapeǤʤΤФ
	(if (not (and uim-xemacs
		      (equal spkey 'escape)))
	    (define-key uim-preedit-map meta-vec 'uim-process-input))


	(setq special-keys (cdr special-keys)))

      (setq mod-keys (cdr mod-keys)))

    ;; ФXEmacsΥץꥨǥåñEscapeϿ
    ;;  ˤޤȤƤȤȤޤʤΤ
    ;;   ǥץꥨǥåñEscapeɬפȤʤ
    ;;   ꤷʤ
    (if (and uim-xemacs
	     (or window-system
		 (and (not window-system)
		      uim-use-single-escape-on-terminal)))
	(define-key uim-preedit-map (vector (list 'escape)) 'uim-process-input))
    ))






;; ޥå׽
(defun uim-init-keymap ()

  (setq uim-mode-map (make-keymap))
  (setq uim-preedit-map (make-keymap))


  ;;; ̾省  032δ֤Control
  (let ((i 0))
    (while (< i 128)
      (if (not (or (keymapp (key-binding (char-to-string i))) ;; Esc/C-x/C-c
		   (= i 21) ;; C-u
		   (= i 7) ;; C-g
		   ))
	  (define-key uim-mode-map (char-to-string i) 'uim-process-input))

	

      ;;; ץꥨǥå
      (if (/= i meta-prefix-char)
	  (define-key uim-preedit-map (char-to-string i) 'uim-process-input))
      (setq i (+ i 1))))



  ;;; ʳControl
  ;;;  Control-91EscapeʤΤǺɤʤ
  ;;; ߥʥǤϤΤؤϰ̵̣
  (let ((i 32))
    (while (< i 127)
      (if (not (and (>= i 97) (<= i 122)))
	  (progn
	    (if (and ;(not (and (>= i 48) (<= i 57)))
		     (not (keymapp (key-binding (vector (list 'control i))))))
		(define-key uim-mode-map (vector (list 'control i)) 'uim-process-input))
	    ;;; ץꥨǥåȤ˴ؤƤ⡤C-91Escapeˤʤä㤦ΤǸ
	    (if (/= i 91)
		(define-key uim-preedit-map (vector (list 'control i)) 'uim-process-input))
	    ))
      (setq i (+ i 1))))


  ;;; meta-prefix-char ǻϤޤ륭 (M- M-C- M-C-ascii)
  (let ((i 32))
    (while (<= i 126)

      (if (/= i meta-prefix-char)
	  (progn
	    ;; keymap˥ޥåפƤʬϿʤ

	    (if (not (keymapp (key-binding (vector (list 'meta i)))))
		(define-key uim-mode-map 
		  (vector (list 'meta i)) 
		  'uim-process-input) )

	    (if (not (keymapp (key-binding (vector (list 'meta 'control i)))))
		(define-key uim-mode-map
		  (vector (list 'meta 'control i)) 
		  'uim-process-input) )
	    
            ;;; uim-preedit-map
	    (define-key uim-preedit-map
	      (vector (list 'meta i))
	      'uim-process-input)

	    ;;; meta contorl-91ϡESC-ESCˤʤ뤬
	    (define-key uim-preedit-map 
	      (vector (list 'meta 'control i)) 
	      'uim-process-input)
	))

      (setq i (+ i 1))))

  ;;;  ü쥭ɲ
  (uim-bind-special-keys)

  ;; Emacsǡߥʥǵưݤ
  ;;  uim-use-single-escape-on-terminaltʤ
  ;;  EscapeñȽɬפȤ뤫ɤ
  (if (and (and uim-emacs (not window-system))
	   uim-use-single-escape-on-terminal)
      (define-key uim-preedit-map (vector (list meta-prefix-char)) 'uim-process-input)
    )

  )

(provide 'uim-keymap)


