;; 
;;  uim.el - Emacs Uim bridge
;;
;;  Copyright (c) 2005, Konosuke Watanabe <nosuke@csc.ne.jp>
;;  All rights reserved.
;;
;;  $Id: uim.el,v 1.20 2005/07/24 21:31:50 nosuke Exp $
;;
;;  Redistribution and use in source and binary forms, with or
;;  without modification, are permitted provided that the
;;  following conditions are met:
;;
;;  1. Redistributions of source code must retain the above
;;     copyright notice, this list of conditions and the
;;     following disclaimer.
;;  2. Redistributions in binary form must reproduce the above
;;     copyright notice, this list of conditions and the
;;     following disclaimer in the documentation and/or other
;;     materials provided with the distribution.
;;  3. Neither the name of authors nor the names of its
;;     contributors may be used to endorse or promote products
;;     derived from this software without specific prior written
;;     permission.
;;
;;  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
;;  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
;;  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
;;  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;;  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
;;  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
;;  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
;;  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;;  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
;;  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
;;  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;


(require 'uim-init)
(require 'uim-debug)
(require 'uim-util)
(require 'uim-key)


;; °˱face֤
(defun uim-select-preedit-face (attr)
  (cond
   ((string-match "s" attr)
    'uim-separator-face)
   ((string-match "ru" attr)
    'uim-preedit-highlight-underline-face)
   ((string-match "r" attr)
    'uim-preedit-highlight-face)
   ((string-match "u" attr)
    'uim-preedit-underline-face)
   (t 
    'uim-preedit-face)))


;; ƥIDμ
(defun uim-get-context-id ()
  (let (id)
    (if (not (null uim-context-id-recycle))
	(progn
	  (setq id (car uim-context-id-recycle))
	  (setq uim-context-id-recycle (cdr uim-context-id-recycle)))
      (setq id uim-context-id-max)
      (setq uim-context-id-max (+ uim-context-id-max 1)))
    id ))


;; ƥIDβ
(defun uim-release-context-id (id)
  (setq uim-context-id-recycle (cons id uim-context-id-recycle)))


(defun uim-check-im (im)
  (if (assoc im (cdr uim-im-list))
      t
    (message (format "uim.el: invalid IM engine: %s" im))))



(defun uim-get-default-engine ()
  (if (and uim-default-im-engine
	   (uim-check-im uim-default-im-engine))
      uim-default-im-engine
    uim-context-default-im-engine))


(defun uim-update-default-engine (engine)

  (when (not (equal uim-context-default-im-engine engine))
    (setq uim-context-default-im-engine engine)
		 
    ;; leimϢ
    ;; default-input-methodȿǤ
    (if (and (featurep 'uim-leim)
	     (boundp 'default-input-method))
	(setq default-input-method
	      (uim-leim-make-im-name
	       uim-context-default-im-engine))))
  )

(defun uim-update-current-engine (engine)
  (uim-debug (format "current: %s" engine))
  (when (not (equal uim-current-im-engine engine))
    (setq uim-current-im-engine engine) 
		 
    ;; leimϢ
    ;; current-input-methodȿǤ
    (if (and (featurep 'uim-leim)
	     (boundp 'current-input-method)
	     current-input-method)

	(setq current-input-method
	      (uim-leim-make-im-name uim-current-im-engine)))))


	
;; ƥȤο
(defun uim-context-create ()
  (let (id)
    ;; IDγ
    (setq id (uim-get-context-id))
    ;; agent¦б륳ƥȤ
    ;;  Uim¦ΥǥեȤ(ƥΤǶ̤Τ)ºŤ
    ;;   LISTͤcarȤ
    ;;   桼ꤷƤ硤ˤͥȤ
    (uim-do-send-recv-cmd 
     (format "%d NEW %s" id (uim-get-uim-code (uim-get-default-engine))))
    id))

;; ƥȤβ
;;  ƥIDϡХåե̤ޤǲʤ
;;  agentǤ⡤ƱХåեǤƱIDѤ
(defun uim-context-delete (id)
  (uim-do-send-recv-cmd (format "%d RELEASE" id))
  (uim-release-context-id id))


;; agentȤ̿ꥢֹ
(defun uim-get-communication-serial-number ()
  (if (= uim-communication-serial-number 65535)
      (setq uim-communication-serial-number 0))
  (setq uim-communication-serial-number
	(+ uim-communication-serial-number 1)))


;; ƥȤΥꥻå
(defun uim-context-reset (id)
  (uim-do-send-recv-cmd (format "%d RESET" id)))

;; ⡼ɥ饤
(defun uim-redraw-mode-line ()
  (force-mode-line-update))


;; ߥ˥Хåեä(뤳Ȥʤ)
(defun uim-minibuffer-enter ()
;  (uim-debug "minibuffer enter")
  nil)

;; ߥ˥ХåեФݤˡߥ˥ХåեuimڤäƤ
(defun uim-minibuffer-exit ()
;  (uim-debug "minibuffer exit")
  (if (and (/= uim-context-id 0) uim-mode)
      (uim-mode-off)))

;; ߥ˥Хåեν˥եå
(add-hook 'minibuffer-setup-hook 'uim-minibuffer-enter)
(add-hook 'minibuffer-exit-hook 'uim-minibuffer-exit)


;; ߤΥɥ饤ѴɽǤ륵ɤȽ
(defun uim-check-candidate-space ()
  ;; ߥ˥Хåեγǡɥuim-candidate-minimum-widthʾ
  (and (= (minibuffer-depth) 0)
       (>= (window-width) uim-candidate-minimum-width)))
       

;; ƥȤνIMѹ
;;  DEFAULTޥɤIM̾򤢤ƤϤʤ
(defun uim-set-context-default (im)
  (uim-do-send-recv-cmd (format "0 DEFAULT %s" im)))


;; Хåեե줿ݤν
(defun uim-focused (id)
  (uim-change-process-encoding)
  (uim-do-send-recv-cmd (format "%d FOCUSED" id)))



;; Хåեե줿ݤν
(defun uim-unfocused (id)
  (uim-do-send-recv-cmd (format "%d UNFOCUSED" id)))


(defun uim-force-off ()
  (uim-context-reset uim-context-id)
  (if uim-leim-active-p
      ;; leimͳǾ夬äƤ硤Ȥ
      (inactivate-input-method)
    ;; ޥʥ⡼ɷͳǾ夬äƤȤ
    (if uim-mode
	(uim-mode-off))))

;; ե򸡽Ф뤿ˡȥХåեڤؤ򸡽
(defun uim-change-buffer ()

  ;;; ХåեΥ꡼ɥ꡼°Ѳ
  (if (and (/= uim-context-id 0)
	   (not (equal uim-buffer-readonly-state buffer-read-only)))
      (progn
;	(uim-debug (format "***read-only mode changed to %s" buffer-read-only))
	(setq uim-buffer-readonly-state buffer-read-only)
	(if buffer-read-only
	    ;; 
	    (uim-force-off))))

  (if (not (eq uim-recent-buffer (current-buffer)))
      (progn
	;; ǶΥХåե򹹿
	(setq uim-recent-buffer (current-buffer))

	;(uim-debug "buffer changed")

	;; ȤꤢХåեѤä餳ã

	;; ˤХåեޤ¸ߤƤե򳰤
	;;   ˥ץưƤ뤳ȤݾڤƤϤ
	(if (and (bufferp uim-focused-buffer)
		 (buffer-name uim-focused-buffer))
	    (save-current-buffer
	      (set-buffer uim-focused-buffer)
	      (uim-unfocused uim-context-id)))
	      

	;; ܤäΥХåե˥ƥȤХե
	(if (and (/= uim-context-id 0)
		 uim-mode)
	    (progn 
	      (uim-focused uim-context-id)
	      ;; ߤͭʥХåե򹹿
	      (setq uim-focused-buffer (current-buffer)))
	  ;; ܤäuim¸ߤǧǤʤäuim-focused-buffernil
	  (setq uim-focused-buffer nil))
	)))

;; ٥ȸ˥եå򤫤뤳ȤǸ
(add-hook 'post-command-hook 'uim-change-buffer)


;; Ѳǽʥ󥸥ΥꥹȤ
(defun uim-get-im-list ()
  (if (not uim-im-list)
      (uim-do-send-recv-cmd (format "0 LIST"))))


;; 󥸥Υ⡼ɤؤ
(defun uim-prop-activate (id proplist)
  (if (> uim-context-id 0)
      (let ((plist proplist))
	(while plist
	  (uim-do-send-recv-cmd (format "%d PROP %s" id (car plist)))
	  (setq plist (cdr plist))))))



(defun uim-change-process-encoding ()
  (uim-debug (format "set-process-encoding to %s" uim-decoding-code))
  (set-process-coding-system 
   uim-agent-process 
   uim-decoding-code
   'iso-8859-1))


(defun uim-get-im-info (im info)
  (cdr (assoc info (cdr (assoc im (cdr uim-im-list ))))))


(defun uim-get-uim-code (im)
  (or (uim-get-im-info im 'uim-code)
      "ISO-8859-1"))

(defun uim-get-emacs-code (im)
  (or (uim-get-im-info im 'emacs-code)
      'iso-8859-1))

(defun uim-get-uim-lang (im)
  (or (uim-get-im-info im 'uim-lang)
      "Unknown"))

(defun uim-get-emacs-lang (im)
  (or (uim-get-im-info im 'emacs-lang)
      "ASCII"))


;; IMѹ
(defun uim-change-im (id im)
  (if (> uim-context-id 0)
      (progn
	(uim-change-process-encoding)
	(uim-do-send-recv-cmd 
	 (format "%d CHANGE %s %s" id im (uim-get-uim-code im)))

	;; labelѤʤΤǤ礦ʤ
	(uim-unfocused id)
	(uim-focused id)

	(if uim-default-im-prop
	    (uim-prop-activate uim-context-id uim-default-im-prop))

	)))



(defun uim-set-encoding (im)
  (let ((uim-code (uim-get-uim-code im)))
    (uim-do-send-recv-cmd 
     (format "0 SETENC %s %s" im uim-code))))


;;; M-x uim-im-switch ǸƤӽФ
;;; leimǤϻȤäƤϤʤ
;;; 
(defun uim-im-switch (&optional im)
  (interactive)
  (when (and (/= uim-context-id 0)
	     (not (featurep 'uim-leim)))
    (if (not im)
	(let (alist)
	  (setq alist (mapcar (lambda (x) (cons (car x) (car x))) (cdr uim-im-list)))
	  (save-window-excursion
	    (setq im (cdr (assoc
			   (completing-read
			    (concat "IM Engine: ")
			    alist nil t)
			   alist) )))))
    (if (uim-check-im im)
	(uim-change-im uim-context-id im)))
  (if (featurep 'uim-leim)
      (message "uim.el: use  M-x set-input-method  when using LEIM")))



;; ꡼ɥ꡼ηٹ
(defun uim-warn-readonly ()
  (message "[uim.el] This buffer is read-only.  You should turn off Uim now."))



;; ⡼ɹԤɲ
(defun uim-modeline-add-propinfo ()
  (if (and uim-emacs
	   (>= emacs-major-version 22))
      ;; for emacs-22
      (setq mode-line-modes 
	    (uim-list-search-append mode-line-modes
				    'minor-mode-alist
				    'uim-prop-string))
    (setq mode-line-format 
	  (uim-list-search-append mode-line-format
				  'minor-mode-alist
				  'uim-prop-string)))
  (setq uim-mode-line-initialized t))

  


;; ⡼ɹԤ
(defun uim-modeline-remove-propinfo ()
  (if (and uim-emacs
	   (>= emacs-major-version 22))
      ;; for emacs-22
      (setq mode-line-modes 
	    (uim-list-search-delete mode-line-modes
				    'uim-prop-string))
    (setq mode-line-format 
	  (uim-list-search-delete mode-line-format
				  'uim-prop-string)))
  (setq uim-mode-line-initialized nil))




;; ХåեλĥƥȤ
(defun uim-context-init ()
  ;; ƥݻХåեꥹȤɲ
  (if (not (memq (buffer-name) uim-using-buffer))
      (setq uim-using-buffer (cons (buffer-name) uim-using-buffer)))

  ;(if uim-default-im-prop
  ;    (uim-prop-activate uim-context-id uim-default-im-prop))

  ;; ٥ιޤʤХ
  (uim-unfocused uim-context-id)
  (uim-focused uim-context-id))
  

;; ХåեΥƥȰʳν
(defun uim-buffer-init ()
  (uim-modeline-add-propinfo)
  (setq uim-buffer-readonly-state buffer-read-only))



;; uimON
(defun uim-mode-on ()
  ;; Хåե꡼ɥ꡼ξ硤ONʤ
  (if buffer-read-only
      (message "This buffer is read-only.")
    (uim-enable-mode-keymap)

    ;; if minor-mode is already acrive, do nothing
    (if uim-mode
	(message "uim.el: uim-mode is already activated.")

      ;; ʳξ
      (when (= uim-context-id 0)
	;; Хåե˥ƥȤʤƽ

	;; ˥ƥȤΥǥեͤ򹹿
;	(if (and uim-default-im-engine
;		 (uim-check-im uim-default-im-engine))
;	    (uim-set-context-default uim-default-im-engine))

	(setq uim-context-id (uim-context-create))
	(uim-context-init)
      (uim-buffer-init))

      ;; ȥƥȤ򤳤ΥХåեΥƥȤˤ
      (uim-focused uim-context-id)

      (setq uim-focused-buffer (current-buffer))
    
      (if uim-default-im-prop
	  (uim-prop-activate uim-context-id uim-default-im-prop))

      ;; Ѵɽ⡼ɤ
      (setq uim-candidate-display-inline uim-candidate-display-inline-default)
      (uim-redraw-mode-line)

      ;; ɽѤΰι
      (setq uim-prop-string uim-prop-string-buffer)

      ;; ޥʡ⡼ͭ
      (setq uim-mode t))))


;; uimOFF
(defun uim-mode-off ()
  
  (if (not uim-mode)
      (message "uim.el: uim-mode is already inactivated.")

    ;; uim-agentݤ˸ƤФ줿硤̿ʤ
    (if uim-agent-process
	(uim-unfocused uim-context-id))
    
    (setq uim-prop-string nil)
    (uim-redraw-mode-line)

    ;; ޥʡ⡼̵
    (setq uim-mode nil)))



;; uimONOFF
(defun uim-mode-switch ()
  (interactive)
  (if uim-mode
      (uim-mode-off)
    (uim-mode-on)))



;; Хåեkill줿饳ƥID
(defun uim-kill-buffer ()
  (if (eq uim-focused-buffer (current-buffer))
      (progn
	;;; ΥХåեľںߤƤ硤ľˤХåեnil
 	(setq uim-focused-buffer nil)))
  (if (/= uim-context-id 0)
      (progn
	;;; ΥХåեuimȤƤ饳ƥȲ
	;;; uim-context-idϥХåեʤΤnil᤹ɬפ̵
	(uim-context-delete uim-context-id)
	;; ƥݻХåե
	(delq (buffer-name) uim-using-buffer)
	)))

;; Хåեkillƻ
(add-hook 'kill-buffer-hook 'uim-kill-buffer)

;; major-modeܤƻ
(add-hook 'change-major-mode-hook 'uim-kill-buffer)



;; uim-agentεư
(defun uim-agent-start ()
  (let (proc
	(buffer (get-buffer-create uim-agent-buffer-name)) )

    ;; ǥ顼ȯƤ⤽¹ԤʤʤΤʿ
    (setq proc (start-process "uim-agent" buffer uim-agent))

    ;(set-process-coding-system proc 'euc-jp 'euc-jp)

    ;; λkill䤤碌ʤ
    (process-kill-without-query proc)

    ;; Хåեꥢ
    (save-current-buffer
      (set-buffer buffer)
      (erase-buffer))

    (setq uim-agent-buffer buffer)

    (set-process-sentinel proc 'uim-process-sentinel)

    ;; ǤȯƤ⡤ץĤ

    proc ))


;; uim-agentȤ
(defun uim-agent-stop (proc)
  (if (and proc
	   (process-status proc))
      (interrupt-process proc)))


;; Ѵ
;;  tabڤ⤳ڤȤ
;;   Ϥɤɽ褦
(defun uim-make-candlist-format (cand count)
  (let (candtmp candstr trunc tab (optstr nil) (candlist '()))
    ;;; ֹ

    (setq candtmp (cdar cand))
    (if (setq tab (string-match "\t" candtmp))
	(setq optstr (substring candtmp (+ tab 1))))
    (setq candstr (substring candtmp 0 tab))

    (if (caar cand)
	(setq candstr (format "%2d.[%s]" count candstr))
      (setq candstr (format "%2d. %s " count candstr)))
    
    ;;; ĹԤ2Ĥˤ櫓
    ;;;  Ǥ⥦ɥ-1ˤ
    ;;;   (ɥƱˤޤ֤Ƥޤ)
    (while (> (uim-string-width candstr) 0)
      (setq trunc (truncate-string-to-width candstr (- (window-width) 1)))
      (setq candlist
	    (append candlist (list (list count
					 trunc
					 (caar cand)
					 optstr ))))

      ;;;  ޤ֤λĤ
      (setq candstr (substring candstr (length trunc)))
      )

    candlist))


;; Ѵ¤ؤꤹ
(defun uim-make-candlist (candidate) 
  (let ( (cand candidate ) (candf '() ) (i 1) )
    ;;; ǽϸʤΤǳ
    (setq candf (list (list 0 (format "%d/%d"(caar cand) (cdar cand)))))
    (setq cand (cdr cand))
    (while cand
      (setq candf (append candf (uim-make-candlist-format cand i)))
      (setq i (+ i 1))
      (setq cand (cdr cand))
      )
    ;;; ǸˤäƤ
    ;;;(setq candf (append (cdr candf) (list (car candf))))

    ;;; ǽˤäƤ
    candf
    ))





;;; Ϳ줿ƥȤ򡤴¸ΥƥȤ˥ޡƲ̤ɽ

(defun uim-insert-recttext (strlist base end vofs)
  (let ( maxwidth preedit phead baseofs ptail  vhead vlbase vlend prebuf
		  offset original-line merge-end (vllist '()) ol)

    ;; ΰ
    (goto-char base)


    ;; ѤΥХ쥤
    (setq ol (make-overlay base (+ base 1)))

    ;; ΤԤιƬ
    (save-excursion
      (beginning-of-line)
      (setq phead (point))
      )

    ;; 񤭤ǸʸΤԤʪ
    (save-excursion
      ;; Ǹend-of-lineƤΤ¿ʬvertical-motionʿ
      (uim-vertical-motion (+ vofs (length strlist)))
      (end-of-line)
      (setq ptail (point)))

    ;; ꥸʥ򤷤Ƥ
    (setq original-line
	  (buffer-substring phead ptail))

    ;; äڡִ(baseΰ֤overlayưפƤ)
    (uim-tab-pad-space phead ptail)

    (setq baseofs (- (- (overlay-end ol) 1) base))

    (delete-overlay ol)
    (setq ol nil)

    (setq base (+ base baseofs))

    (goto-char base)


    ;; ƴβ۹Ƭ֤
    (save-excursion
      (uim-vertical-motion 0) ;; ⤦vertical-motionǤ¿ʬס
      (setq vhead (point)))


    ;; ƬľޤǤʸɽ
    (setq offset (uim-string-width (buffer-substring vhead base)))

    ;; ˴ꤹʤ硤Ĵ
    ;; ϤɤƱ(strlistθġtab̵Ϥ)
    (setq maxwidth (uim-string-width (car strlist)))

    ;; ƬΥ +  >= ɥ ξ硤ƬΥĴ
    (if (>= (+ offset maxwidth) (window-width))
	(setq offset (- (- (window-width) 1) maxwidth)))


    ;; offsetͤѤƥޡ
    (save-excursion
      (let (mergeline (strl strlist) )

	(uim-vertical-motion vofs)

	;; ֤ɤˤ뤫(βʬޡɬפ뤫)
	;;   0ξ硤ϲ۹ԤХåեǸβ۹ԤȤȤˤʤ
	(save-excursion
	  (setq mergeline (uim-vertical-motion (length strl))))


	;; ޡ
	(let ( (i 1) 
	       vlhead linetmp padding overflow )

	  ;(fmt (format "%%s%%%ds" maxwidth ))

	  ;; ޤ¸Υǡ񤭤ʬ˴ؤ
	  (while (<= i mergeline)

	    (uim-vertical-motion 1)

	    ;; ʲ۹Ƭɸ
	    (setq vlhead (point))

	    ;; 񤭥ǡ
	    (save-excursion
	      ;; ʪɸ
	      (end-of-line)
    
	      ;; Ǥ⥦ɥ
	      (setq linetmp 
		    (truncate-string-to-width 
		     (buffer-substring vlhead (point))
		     (window-width)))


	      ;; ƬoffsetʬpaddingȤƳ(­϶)
	      (setq padding
		    (truncate-string-to-width linetmp offset nil 32))

	      (setq vllist (append vllist (list (+ vlhead (length padding)))))

	      ;; 줿ʬ
	      (if (>= (uim-string-width linetmp) (+ offset maxwidth) )
		  (setq overflow 
			(truncate-string-to-width linetmp 
						  (uim-string-width linetmp)
						  (+ offset maxwidth) 32))
		(setq overflow ""))

	      ;; ºݤ˾
	      (save-excursion
		;; ޤ۹Ƭ鲾۹ޤǾä
		(goto-char vlhead)

		(delete-region vlhead (+ vlhead (length linetmp)))
		;; ƥȤpaddingoverflow
		;(insert (concat (format fmt padding (car strl)) overflow)))
		(insert (concat padding 
				(uim-format-string (car strl) maxwidth)
				overflow)))

		
	      ;; ޡιϿ(ޡκǸΤ˻Ȥ)
	      (save-excursion
		(end-of-line)
		(setq merge-end (point)))
	      
	      (setq strl (cdr strl))
	      (setq i (+ i 1))
	      
	      )))


	;; ɲʬ
	(let (  vlhead )

	  ;(fmt (format "\n%%%ds%%s" offset) )

	  (while strl

	    ;; Хåե
	    (goto-char (point-max))
	    ;(end-of-buffer)

	    ;; β۹Ƭ
	    (setq vlhead (+ (point) 1))

	    ;; 
	    ;(insert (format fmt "" (car strl)))
	    (insert (concat "\n"
			    (make-string offset 32)
			    (car strl)))
			    
	    
	    (setq vllist (append vllist (list (+ vlhead offset))))

	    ;; ˹ذưޡʬι򹹿
	    (end-of-line)
	    (setq merge-end (point))

	    (setq strl (cdr strl))))
	))

 
    (force-mode-line-update)

    ;; ᤹Ȥϡpheadmerge-endޤǾäơoriginal-lineinsert褤
    (list original-line phead merge-end vllist)
    ))




;; Ϳ줿ƹԤ򡤺ιԤˤ碌ƻͳѤ
(defun uim-make-recttext (candidate &optional align max)

  (let ((maxwidth 0) fmt (candidate-f '()) )

    (if (not max)
	;; ʸƥȤκ"ɽ"
	(let ( width (cand candidate) )
	  (while cand
	    (setq width (uim-string-width (car cand)))
	    (if (> width maxwidth)
		(setq maxwidth width))
	    (setq cand (cdr cand))))
      (setq maxwidth max))

    (let ( (cand candidate) )
      (while cand
	(setq candidate-f
	      (append candidate-f
		      (list 
		       ;(format fmt (car cand))
		       (uim-format-string (car cand) maxwidth align)
		       )))
	(setq cand (cdr cand))))

    candidate-f
    ))



;; uim-insert-recttextƤäơ᤹
(defun uim-remove-recttext (original)
  ;; 򤷤Τ褵
  (let ( (original-line (car original))
	 (start (nth 1 original))
	 (end (nth 2 original))
	 )


    (delete-region start end)


    (save-excursion
      (goto-char start)
      (insert original-line))
    )
)



;; Ѵ(ɬץꥨǥåȤ¸ߤ)
(defun uim-insert-candidate (candidate base end)
  (let (candf maxwidth rect)
    ;; ҤȤޤֹĤĹ·ꥹȤѴ
    (setq candf (uim-make-candlist candidate))

    ;; 򤳤Ƿ׻
    (setq maxwidth
	  ((lambda(l)
	     (let ((max 0))
	       (while l
		 (if (> (car l) max)
		     (setq max (car l)))
		 (setq l (cdr l)))
	       max))
	   (mapcar (lambda (x) (if (stringp (cadr x)) (uim-string-width (cadr x)) 0)) candf)))
    

    (if uim-candidate-display-inline
	;; 饤ɽ

	(progn
	  ;; ƬϸοʤΤǡƺǸ˲
	  (setq candf
		(append (cdr candf)
			(list (list (caar candf)
				    (format (format "%%%ds" maxwidth)
					    (car (cdar candf)))))))

	  ;; Ĥġꥸʥ
	  (setq rect (uim-make-recttext (mapcar 'cadr candf) nil maxwidth))


	  (setq uim-candidate-original (uim-insert-recttext rect base end uim-candidate-ofs))

	  ;; 䤬ϤߤǤƤ⥦ɥΰ֤򤺤餵ʤ
	  (if uim-window-start
	      (set-window-start nil uim-window-start nil))


 	  (let ( face (i 1) 
		      (vllist (nth 3 uim-candidate-original)))


 	    (while vllist
	      (cond 
	       ((cadr (cdar candf))
		(setq face 'uim-candidate-selected-face))
	       ((= (caar candf) 0)
		(setq face 'uim-candidate-nth-face))
	       ((= (% (caar candf) 2) 1)
		(setq face 'uim-candidate-odd-face))
	       ((= (% (caar candf) 2) 0)
		(setq face 'uim-candidate-even-face)))

	      ;; Х쥤դ
	      (let (ol)
		(setq ol (make-overlay (car vllist) 
				       (+ (car vllist) (length (car rect)))))
		(overlay-put ol 'face face)	
		(overlay-put ol 'priority 10)
		(setq uim-candidate-overlays (cons ol uim-candidate-overlays)))

	      (setq vllist (cdr vllist))
	      (setq rect (cdr rect))
	      (setq candf (cdr candf)))

	    ))

      ;; ߥ˥Хåե˽Ϥ(ʤˤδñ)
      (let ((out "")  )
	(while candf
	  (setq out (concat out (car (cdar candf ))))
	  (setq candf (cdr candf)))
	(message out)
	)
      )))



;; Ѵä
(defun uim-remove-candidate ()
  (if uim-candidate-display-inline
      (progn

	(uim-remove-recttext uim-candidate-original)

	;; Х쥤ä
	(mapcar 'delete-overlay uim-candidate-overlays)
	(setq uim-candidate-overlays nil)

	;; 䤬ϤߤǤƤ⥦ɥΰ֤򤺤餵ʤ
	(if uim-window-start
	    (set-window-start nil uim-window-start nil))

	)
    (message nil))
  )


;;; Х쥤γϰ֤ʸä顤˥Х쥤ƨ
(defun uim-preedit-overlay-insert-head (ol after head tail &optional len)
  (if after
      (move-overlay ol (+ (overlay-start ol) (- tail head))
		    (overlay-end ol))))

;; ץꥨǥå
(defun uim-insert-preedit ( preedit base )
  ;; preeditagentϤ줿preeditꥹ

    (setq uim-preedit-cursor nil)


    (let (lbase window-head window-tail window-buffer)
      (setq lbase (point))

      ;; λǤΥɥξüϿ
      (setq uim-window-start (window-start))

      ;; ѹԲǽʬѹǽˤ
      ;;  freezeˤ碌ȥߥåȤǤ㤦
      (setq uim-inhibit-read-only inhibit-read-only)
      (setq inhibit-read-only t)

      (goto-char lbase)

      (set-window-start nil uim-window-start nil)

      )


    (save-excursion
      (setq uim-candidate-start nil)
      (setq uim-candidate-ofs 0)
      (let ( (pe preedit) preedit-start ol )
	(setq uim-preedit-start base)

	(while pe

	  (setq preedit-start (point))

	  ;; (˸ɽ)
	  (if (string-match "c" (format "%s" (caar pe)))
	      (progn
		(setq uim-candidate-start (point))
		))


	  
	  ;; ºݤʸ
	  (if (> (length (car (cdar pe))) 0)
	      (progn

		(insert (car (cdar pe)) )

		;; Х쥤դ
		(let (ol)
		  (setq ol (make-overlay preedit-start (point)))
		  (overlay-put ol 'face (uim-select-preedit-face (format "%s" (caar pe))))
		  (overlay-put ol 'priority 10)
	      ;;; դʤȡץꥨǥåȤޤ֤֤Ϥޤˤ
		  (if (not uim-xemacs)
		      (overlay-put ol 'insert-in-front-hooks '(uim-preedit-overlay-insert-head)))
		  (setq uim-preedit-overlays (cons ol uim-preedit-overlays)))
		
		))

	  (if (string-match "c" (format "%s" (caar pe)))
	      (setq uim-preedit-cursor (point)))

	  (setq uim-preedit-end (point))

           ;;; uim-candidate-start

	  (setq pe (cdr pe)))))


    (if (and uim-preedit-cursor
	     (> uim-preedit-cursor (point-min)))
	;; ץꥨǥåȤȤȤϾʤȤ1ʸ¸
	(progn
	  (setq uim-preedit-cursor-ol (make-overlay (- uim-preedit-cursor 1)
						    uim-preedit-cursor ))
	  ))



    (if (not uim-candidate-start)
	(progn
	  (setq uim-candidate-start uim-preedit-start)
	  (setq uim-candidate-ofs 
		(uim-vertical-distance uim-candidate-start uim-preedit-end))
	  )
      (setq uim-candidate-ofs 
	    (uim-vertical-distance uim-candidate-start uim-preedit-end)))
    
    )




;; ץꥨǥåȺ
(defun uim-remove-preedit ()

  (save-excursion
    ;; Х쥤ű
    (mapcar 'delete-overlay uim-preedit-overlays)
    (setq uim-preedit-overlays nil)

     (if uim-preedit-cursor-ol
 	(progn
 	  (delete-overlay uim-preedit-cursor-ol)
 	  (setq uim-preedit-cursor-ol nil)))

    ;; ץꥨǥåȤΤΤä
    (delete-region uim-preedit-start uim-preedit-end))
  

  ;; ץꥨǥåȤä뤿Ӥѹǽե饰᤹
  (setq inhibit-read-only uim-inhibit-read-only))







;; undoꥹȲ
(defun uim-enable-undo ()
  (if uim-buffer-undo-list-saved
      (progn
	(buffer-enable-undo)
	(setq buffer-undo-list uim-buffer-undo-list)
	(setq uim-buffer-undo-list-saved nil) 
	)))


;; undoꥹ
(defun uim-disable-undo ()
  (if (not uim-buffer-undo-list-saved)
      (progn
	(setq uim-buffer-undo-list buffer-undo-list)
	(buffer-disable-undo)
	(setq uim-buffer-undo-list-saved t))))




;; uim-agent֤Ƥ
(defun uim-process-agent-output (str)
  (let ( preedit preedit-pre candidate candidate-pre oldcandidate)

    (let (modified)
      (setq modified (buffer-modified-p))  ;;; ѹե饰

      ;; ɽʤ餤ä(ˤ餺Ȥˤپä)
      (if uim-candidate-exist
	  (progn
	    (uim-remove-candidate)
	    (setq candidate-pre t)
	    (setq uim-candidate-exist nil)))

      ;;; Ѵ⥯ꥢ
      (setq uim-candidate nil)
    
      ;; ץꥨǥåȤɽʤ餤ä
      (if uim-preedit-exist
	  (progn
	    (goto-char uim-preedit-start)
	    (uim-remove-preedit)
	    (setq preedit-pre t)
	    (setq uim-preedit-exist nil)
	    ))
      ;;; ץꥨǥåȤäƤ
      (setq uim-preedit nil)


      (set-buffer-modified-p modified) ;; ѹե饰
      )

    (uim-debug (format "%s" str))

    (while str
      (cond ((string= (caar str) "n")
	     
	     (let ( keyvec )
	       (setq keyvec (car (cdar str)))
	       (if keyvec
		   ;; ̤ξ
		   (progn
		     (uim-process-keyvec keyvec)
		     (undo-boundary) ) ;; C-gλˤƤФʤ
		 ;; nʤɽä

		 )))

			 
	    ;; ߥå
	    ((string= (caar str) "s")
	     (uim-enable-undo)
	     (setq uim-candidate nil)
	     (insert (car (cdar str)))
	     (undo-boundary) 
	     ;; Ǹ˥ޥåפ
	     (uim-enable-mode-keymap))
	    

	    ;; Ѵ
	    ((string= (caar str) "c")
	     (setq uim-candidate (cdar str) ))

	    ;; ץꥨǥå
	    ((string= (caar str) "p")
	     (uim-enable-preedit-keymap) ;; ޥåѹ
	     (setq uim-preedit ( cdar str)))
	    
	    ;; ꥢ 
	    ((string= (caar str) "e")
	     (setq uim-preedit nil)
	     (uim-enable-mode-keymap)	
	     (uim-enable-undo))

	    ;; label
	    ((string= (caar str) "l")
	     (let (mode1 mode2 engine default)

	       (setq engine (car (nth 1 (car str))) )

	       ;; agentδǥեȤΥ󥸥
	       (setq default (cdr (nth 1 (car str))) )

	       ;; agent֤Ƥǥեȥ󥸥̾Ͽ
	       (uim-update-default-engine default)

	       ;; agent֤ƤߤΥ󥸥̾Ͽ
	       (uim-update-current-engine engine)

	       (setq mode1 (car (nth 2 (car str))))
	       (setq mode2 (car (nth 3 (car str))))

	       (let ((code
		      (uim-get-emacs-code uim-current-im-engine)))

		 (if (not (equal code uim-decoding-code))
		     (progn
		       (setq uim-decoding-code code)
		       (uim-change-process-encoding))))
		     

	       (setq uim-current-prop-label
		     (if mode2
			 (format "%s%s" mode1 mode2)
		       (setq mode1 (format "%s" mode1))))


	       ;; ⡼ɥ饤ʸ򹹿
	       (setq uim-prop-string-buffer
		     (format " %s[%s]"
			     uim-current-im-engine 
			     uim-current-prop-label))
	       (when uim-mode
		 (setq uim-prop-string uim-prop-string-buffer)
		 (uim-redraw-mode-line))
	       )
	     )

	    
	    ((string= (caar str) "L")
	     (let ( (default (car (cdar str))) )

	       ;; agent֤Ƥǥեȥ󥸥̾Ͽ
	       (uim-update-default-engine default)

	       (setq uim-im-list 
		     (cons default
			   (mapcar (lambda (x) (cons (car x)
						     (cdr (or (assoc (nth 2 x) uim-lang-code-alist)
							      (assoc "Other" uim-lang-code-alist)))))
				   (cdr (cdar str)))))
	       )
	     )

	    ((string= (caar str) "a")
	     (setq uim-command-result t))

	    ((string= (caar str) "f")
	     (setq uim-command-result nil))

	    (t 
	     nil
	     ;(uim-debug "undefined message")
	     )
	    )
      
      (setq str (cdr str)))




    ;; ץꥨǥåȤ褬ɬפʤ
    (if uim-preedit
	(let (modified)

	  (uim-disable-undo)

	  (setq modified (buffer-modified-p))  ;;; ѹե饰

	  (setq uim-preedit-exist t)


	  ;; äޤǥץꥨǥåȤ̵ä
	  (if (and (not preedit-pre)
		   (= (minibuffer-depth) 0))
	      ;; 뤹
	      (progn
		(uim-freeze-the-world)))

	  (uim-insert-preedit uim-preedit (point))

	  ;;; ɽ٤Ѵ䤬äơɽǤ륦ɥˤ
	  (if (and uim-candidate
		   (uim-check-candidate-space))
	      (progn
		;; ɽ
		(setq uim-candidate-exist t)
		(uim-insert-candidate uim-candidate uim-candidate-start uim-preedit-end)))
	  (set-buffer-modified-p modified) ;; ѹե饰
	  )
      
      ;; ϥץꥨǥåȤʤơäޤǥץꥨǥåȤ¸ߤƤ
      (if (and preedit-pre
	       (= (minibuffer-depth) 0))
	  (progn
	    ;; 
	    (uim-unfreeze-the-world)
	    ;; ĤǤ˥ݽ(ĤΤǤҤȤޤߤƤߤ)
	    ;(uim-update-window)
	    )))


    ;; ư
    (if uim-preedit-exist
	(progn
	  (if (not uim-preedit-cursor)
	      (setq uim-preedit-cursor uim-preedit-end))
	  (goto-char uim-preedit-cursor)))
    
    (if uim-preedit-cursor-ol
	(goto-char (overlay-end uim-preedit-cursor-ol)))

    ))





;; ֥ޥɽ
(defun uim-funcall (bind arg)
  (uim-debug (format "uim-funcall: %s %s" bind arg))
  (if bind
      (cond ((and (or (eq bind 'next-line)
		      (eq bind 'previous-line))
		  (not arg))
	     (funcall bind 1))
	    ((and (eq bind 'self-insert-command)
		  (eq arg nil))
	     ;;  ХåեΥե뤬åƤñȯʸб
	     (funcall bind 1))
	    ((eq bind 'just-one-space)
	     ;;  ХåեեåMeta spaceˤб
	     (funcall bind))
	    (t 
	     (funcall bind arg)))
    nil))



;; ¾ΥץˤäƥХåեåƤ뤫ɤȽ
(defun uim-buffer-locked-p ()
  (and (buffer-file-name)
       (file-locked-p (buffer-file-name))
       (not (eq t (file-locked-p (buffer-file-name))))))



;; Ϥ˴ϢŤ줿ؿ
;;  ƥȡδؿƤФ
(defun uim-process-input (&optional arg)
  (interactive "*")

  (if (and (not current-prefix-arg)
	   (not (uim-buffer-locked-p)))

      (progn
	;; ֥ޥɰ̵
	;;  ºݤ˥Ϥuim-agentϤ 
	;; (Ȥ1ȥ or Esc + 1ȥˤޥåפƤʤΤ)
	(let (sendkey keyvec newvec)
	  (if uim-xemacs
	      (progn
		(setq keyvec (this-command-keys))
		(setq sendkey (key-description keyvec))
		(if (= (length sendkey) 1)
		    (setq sendkey (format "X-%s" sendkey)))
	      
		(setq sendkey (format "[%s]" sendkey)))
	    (setq keyvec (this-command-keys-vector))
	    ;; Emacsξ硤function-key-mapѴƤ
	    (if (vectorp (setq newvec 
			       (cdr (assq (aref keyvec 0) function-key-map))))
		(setq keyvec newvec))
	    ;; [127]ñȯä[8](C-h)֤ƤΤ
	    (if (equal keyvec [127])
		(setq keyvec [8]))

	    (if (equal keyvec [27 27])
		(setq keyvec [27]))

	    (setq sendkey keyvec))

	  (uim-do-send-recv-cmd 
	   (format "%d %s" uim-context-id sendkey)))
	)

    
    ;; ֥ޥɰ硤̤
    (let (bind keyvec)
      
      (if current-prefix-arg
	  (uim-debug (format "current-prefix-arg: %s" current-prefix-arg)))

      (if (uim-buffer-locked-p)
	  (uim-debug (format "buffer locked")))

      (if uim-xemacs
	  (setq keyvec (this-command-keys))
	(setq keyvec (this-command-keys-vector)))
      (setq bind (uim-getbind keyvec))
      (if (commandp bind)
	  (uim-funcall bind current-prefix-arg)))))


;;; ɥꥵȤȤ˸ƤФ
;;;  ɥե줿ȤˤƤӽФȾ褤
;;; Ѵ򱣤
(defun uim-window-changed ()
  (let (candidate)
  
    (if (and uim-candidate-exist
	     uim-candidate-display-inline)
	(progn
	  (setq candidate uim-candidate)
	;     (uim-debug "*** uim-window-changed")
       ;; Ȥꤢɽää
	  (uim-do-send-recv-cmd (format "%d HIDE" uim-context-id))

      ;; Ȥȸ䤬¸ߤơ˾꤬Фޤɽ
      ;;  uim-candidateHIDE˾ä㤦Τ
	  (if (and candidate
		   (uim-check-candidate-space))
	      (uim-do-send-recv-cmd (format "%d SHOW" uim-context-id)))
	  
	  ))))

  



(add-hook 'window-configuration-change-hook 
	  'uim-window-changed )
  


;; agentˡƤuimѤƤХåեΥ
;; ѿ
(defun uim-init-all-local-var ()
  (let ((local-var uim-local-var))
    (while local-var
      (set (caar local-var) (cdar local-var))
      (setq local-var (cdr local-var)))))

;; ץθĥ
;;   sleepƤȤϤɤ褦
(defun uim-process-sentinel (proc stat)
  (message "%s" stat)

  (setq uim-agent-process nil)

  ;;; Ѵɽ˻Ȥ
  (let ((old-buffer (current-buffer)))
    (while uim-using-buffer
      (set-buffer (car uim-using-buffer))
      ;; ɽä˾ä
      (uim-process-agent-output '(("n")))
      ;; ХåեǤΥ⡼ɤڤ
      (if uim-mode
	  (uim-mode-off))
      ;; ΥХåեƤΥѿ
      (uim-init-all-local-var)
      ;; ⡼ɥ饤󤫤ä(ɤäˤޤȤ᤿Τ⡩)
      (delq 'uim-prop-string (cdr mode-line-format))
      ;; ꥹȤä
      (setq uim-using-buffer (cdr uim-using-buffer))
      )))



;; uim-agentƤʤå
(defun uim-check-agent ()
  (if (not uim-agent-process)
      (setq uim-agent-process (uim-agent-start))))


;; ޥäƷ̤Ԥ
(defun uim-do-send-recv-cmd (cmd)
  (let ( (serial (uim-get-communication-serial-number)) )
    (setq uim-command-result nil)
    (uim-send-cmd cmd serial)
    ;; always block
    (uim-wait-recv serial)))


;; ޥɤuim-agent
(defun uim-send-cmd (cmd serial)

  (uim-debug (concat (format "%s " serial) cmd ))

  ;; uim-agent뤫ɤå
  (uim-check-agent)

  (save-current-buffer
    (set-buffer uim-agent-buffer)
    (erase-buffer))

  (process-send-string uim-agent-process
		       (concat (format "%s " serial) cmd "\n")))


;; ХåեbufƬ1Ԥɤ߽Ф
(defun uim-get-oneline-from-buffer (buf)
  (save-current-buffer
    (set-buffer buf)
    (when (> (buffer-size) 0)
      (goto-char 1)
      (end-of-line)
      (if (looking-at "\n")
	  (buffer-substring 1 (point))
	nil))))

;; ХåեbufƬ1Ԥä
(defun uim-delete-oneline-from-buffer (buf)
  (save-current-buffer
    (set-buffer buf)
    (delete-region (progn (beginning-of-line) (point))
		   (progn (end-of-line) (point)))))

;; ̤Ԥʬ
;;  ׵˥ꥢֹդơǥޥå󥰤
;;   ޥå󥰤פʤäɤФƼԤ
;;  δؿϤΤϡꥢֹʬȤ
(defun uim-wait-recv(serial)
  (let (rstr exist accept start )

    ;;; ǽϤԤ
    (setq accept uim-agent-accept-timeout)
    (setq start (nth 1 (current-time))) 

    (catch 'read-loop
      (while (and (process-status uim-agent-process)
		  (< (abs (- start (nth 1 (current-time))))
		      uim-agent-timeout))
		  
	(accept-process-output uim-agent-process 0 accept)

	(if (setq rstr (uim-get-oneline-from-buffer uim-agent-buffer))
	    (if (= (car (read rstr)) serial)
		(throw 'read-loop t)
	      (uim-delete-oneline-from-buffer uim-agent-buffer)))

	;; Ԥľ硤acceptΥॢȤޤǤλ֤ܤˤ(̵̣)
	(setq accept (* accept 2))))

    (cond ((not (process-status uim-agent-process))
	   ;; ץ
	   (message "uim.el: state of uim-agent has been changed (die?)"))
	  ((not rstr)
	   (message "uim.el: ***TIMEOUT*** uim-agent did not respond")
	   ;; ٥ॢȤ顤agent̵Τ褤
	   ;;  ٤⥿ॢȤ򤹤Τɤ
	   ;; ȤƤСϾsentinelƤ 
	   (uim-agent-stop uim-agent-process))
	  (t
	   ;; ʳǤΥ...Ǥ󤫡ʡ
	   (uim-process-agent-output (cddr (read rstr)))
	   )
	  )))




(defun uim-enable-preedit-keymap ()
  (setcdr (assq 'uim-mode minor-mode-map-alist) 
	  uim-preedit-map) )

(defun uim-enable-mode-keymap ()
  (setcdr (assq 'uim-mode minor-mode-map-alist) 
	  uim-mode-map))

(defun uim-disable-mode-keymap ()
  (setcdr (assq 'uim-mode minor-mode-map-alist) 
	  nil) )


(defun uim-disable-keymap-temporally ()
  (setq uim-keymap-temp
	(cdr (assq 'uim-mode minor-mode-map-alist)))
  (setcdr (assq 'uim-mode minor-mode-map-alist) nil))

(defun uim-enable-keymap-temporally ()
  (setcdr (assq 'uim-mode minor-mode-map-alist) uim-keymap-temp)
  (setq uim-keymap-temp nil))
  



(defun uim-get-face-list (start end)
  (let ((facelist '()) tail face)
    (catch 'face-loop
      (while t
	(setq tail
	      (next-single-property-change start 'face (current-buffer) end))
	(if (setq face (get-text-property start 'face))
	    (setq uim-facelist (cons (list face start tail) uim-facelist)))
	(setq start tail)
	(if (= start end)
	    (throw 'face-loop t))))
    ))




(defun uim-get-facelist-window (win)
  (select-window win)
  (save-excursion
    (uim-get-face-list (window-start) (window-end nil t))))


(defun uim-get-facelist-buffer ()
  ;; ƱХåեΤ륦ɥФƽ
  ;;;  save-window-excursion ȤäƤޤ
  ;;;  ɤ⤳δؿƤФ٤˥ɥѤäƤޤ
  ;;;  ꤬롩
  (save-selected-window
    (mapcar 'uim-get-facelist-window
	    (get-buffer-window-list (current-buffer) t 'visible))
    uim-facelist))



;; font-lock-modeɤ߹ޤƤΤ߸ƤФ
(defun uim-disable-font-lock-mode ()
  ;; save and disable font-lock-mode
  (setq uim-font-lock-mode font-lock-mode)
  (setq uim-font-lock-verbose font-lock-verbose)
  (when font-lock-mode
    (font-lock-mode 0)
    (setq font-lock-verbose nil)))


;; font-lock-modeξΤ߸ƤФ
(defun uim-enable-font-lock-mode ()
  (when uim-font-lock-mode
    (font-lock-mode uim-font-lock-mode)
    (setq font-lock-verbose uim-font-lock-verbose)))



(defun uim-put-face (facelist)
  (put-text-property (nth 1 facelist) (nth 2 facelist)
		     'face (car facelist)))


(defun uim-freeze-the-world ()
  (let (facelists)
    ;; Ⱦ˥ХåեƬޤǤоݤȤʤäƤޤ
    ;;  Хåե򥹥뤷ʤȤȤǤȤ
    ;;  ƤΥɥˤĴ
    (setq uim-facelist nil)
    (uim-get-facelist-buffer)
    ;(setq facelists (uim-get-face-list 1 (point-max)))
    (if (memq 'font-lock features)	     
	;;Ȥfont-lock-modeѤƤΤ
	(progn
	  (uim-disable-font-lock-mode)
	  (mapcar 'uim-put-face uim-facelist)))
    ))



;; ᤹
;;  ɤФΤ狼󤱤ɤȤꤢfont-lock᤹
(defun uim-unfreeze-the-world ()
  (let (facelists)
    (if (memq 'font-lock features)
	(uim-enable-font-lock-mode))
    ))


;; Ѵɽ⡼ɤؤ
(defun uim-switch-candidate-display-mode ()
  (interactive)

  (let (candidate-exist)

    ;; 䤬¸ߤƤ뤫ФƤ
    (setq candidate-exist (and uim-candidate-exist))

    ;; ̤˻ȤäƤ硤ɽΥ⡼ؤϤꤨʤϤ
    ;; Ȥꤢ¸ɽϾä
    (if candidate-exist
	(uim-do-send-recv-cmd (format "%d HIDE" uim-context-id)))

    (setq uim-candidate-display-inline
	  (not uim-candidate-display-inline))

    ;; ⡼ؤ塤äޤǸ䤬äʤ顤ޤɽ
    (if candidate-exist
	(uim-do-send-recv-cmd (format "%d SHOW" uim-context-id))))
  
  (uim-update-window)
  )


;; agentϢν
(uim-init-agent)

(provide 'uim)

